/* b_matrix_io.c                                            */
/*                                                          */
/* Andrew Janke - rotor@cmr.uq.edu.au                       */
/*                                                          */
/* Largely nicked from input_transform in gen_xf_io.c       */
/*   in volume_io in an attempt at being compatible         */
/*                                                          */
/* Wed Apr  4 14:38:31 EST 2001 - initial version           */

#include <volume_io.h>
#include "b_matrix_io.h"
#define MAX_B_VALS  1000

extern int verbose;

static const STRING B_MATRIX_FILE_HEADER = "MNI Diffusion B_Matrix File";
static const STRING B_MATRIX_TYPE = "B_Matrix_Type";
static const STRING NORMAL_B_MATRIX = "Normal_B_Matrix";
static const STRING B_MATRIX = "B_Matrix";

/* returns a new B_Matrix struct (pointer)                */
B_Matrix *new_b_matrix(int ncoeff, int nbvals)
{
   int      i;
   B_Matrix *tmp;

   if(nbvals > MAX_B_VALS){
      fprintf(stderr, "Maximum # of bvals exceeded (%d), using default\n", nbvals);
      }

   ALLOC_VAR_SIZED_STRUCT(tmp, Real, 10);
   tmp->ncoeff = ncoeff;
   tmp->nbvals = nbvals;

   /* allocate for the array */
   SET_ARRAY_SIZE(tmp->B, 0, nbvals, 10);
   for(i = 0; i < nbvals; i++){
      ALLOC(tmp->B[i], ncoeff);
      }

   return tmp;
   }

/* reads in a B_Matrix from a file                        */
Status input_b_matrix(const char *bmat_file, B_Matrix * Bmat)
{
   int      i, j;

   STRING   line;
   STRING   type_name;
   STRING   str;
   Real     value;
   FILE    *file;

   /* parameter checking */
   if(bmat_file == NULL){
      print_error("input_b_matrix(): passed NULL FILE.\n");
      return (ERROR);
      }

   file = fopen(bmat_file, "r");
   if(file == NULL){
      print_error("input_b_matrix(): error opening B_Matrix file.\n");
      return (ERROR);
      }

   /* okay read the header */
   if(mni_input_string(file, &line, (char)0, (char)0) != OK){
      delete_string(line);
      print_error("input_b_matrix(): could not read header in file.\n");
      return (ERROR);
      }

   if(!equal_strings(line, B_MATRIX_FILE_HEADER)){
      delete_string(line);
      print_error("input_b_matrix(): invalid header in file.\n");
      return (ERROR);
      }

   /* --- read the type of B_Matrix */
   if(mni_input_keyword_and_equal_sign(file, B_MATRIX_TYPE, FALSE) != OK)
      return (ERROR);

   if(mni_input_string(file, &type_name, (char)';', (char)0) != OK){
      print_error("input_b_matrix(): missing b_matrix type.\n");
      return (ERROR);
      }

   if(mni_skip_expected_character(file, (char)';') != OK)
      return (ERROR);

   if(!equal_strings(type_name, NORMAL_B_MATRIX)){
      print_error("input_b_matrix(): invalid b_matrix type.\n");
      delete_string(type_name);
      return (ERROR);
      }
   delete_string(type_name);

   /* --- read the next string */
   if(mni_input_string(file, &str, (char)'=', (char)0) != OK)
      return (ERROR);

   if(!equal_strings(str, B_MATRIX)){
      print_error("Expected %s =\n", B_MATRIX);
      delete_string(str);
      return (ERROR);
      }
   delete_string(str);

   if(mni_skip_expected_character(file, (char)'=') != OK)
      return (ERROR);

   /* now read the lines of the b_matrix */
   if(verbose)
      fprintf(stderr, "Reading B_Matrix:");
   for(i = 0; i < MAX_B_VALS; i++){

      /* allocate a bit ot memory */
      SET_ARRAY_SIZE(Bmat->B, Bmat->nbvals, Bmat->nbvals + 1, 10);
      ALLOC(Bmat->B[i], Bmat->ncoeff);

      Bmat->nbvals++;

      /* get a row */
      for(j = 0; j < Bmat->ncoeff; j++){
         if(mni_input_real(file, &value) == OK){
            Bmat->B[i][j] = value;
            }
         else{
            if(mni_skip_expected_character(file, (char)';') == OK){
               Bmat->nbvals = i;
               if(verbose)
                  fprintf(stderr, " %dx%d B Values Read\n", i, Bmat->nbvals);
               return (OK);
               }
            else{
               print_error("input_b_matrix(): error reading b_matrix [%d,%d]\n", i + 1,
                           j + 1);
               return (ERROR);
               }
            }
         }

      if(verbose){
         fprintf(stderr, ".");
         fflush(stderr);
         }
      }

   /* SHOLDN'T BE REACHED */
   print_error("input_b_matrix(): Glark! Something is amiss in the State of Kansas\n");
   return (ERROR);
   }
