/* mincdti.c                                                                  */
/*                                                                            */
/* The MINC diffusion calculator...(with EPI distortion calculation)          */
/*                                                                            */
/* Andrew Janke - rotor@cmr.uq.edu.au                                         */
/* Mark Griffin - mark.griffin@cmr.uq.edu.au                                  */
/* Center for Magnetic Resonance                                              */
/* The University of Queensland                                               */
/* http://www.cmr.uq.edu.au/~rotor                                            */
/*                                                                            */
/* Copyright (C) 2003 Andrew Janke and Mark Griffin                           */
/* This program is free software; you can redistribute it and/or              */
/* modify it under the terms of the GNU General Public License                */
/* as published by the Free Software Foundation; either version 2             */
/* of the License, or (at your option) any later version.                     */
/*                                                                            */
/* This program is distributed in the hope that it will be useful,            */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              */
/* GNU General Public License for more details.                               */
/*                                                                            */
/* You should have received a copy of the GNU General Public License          */
/* along with this program; if not, write to the Free Software                */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <float.h>
#include <math.h>
#include <ParseArgv.h>
#include <voxel_loop.h>
#include <time_stamp.h>
#include <volume_io.h>
/* someone fix volume_io! */
#undef X
#undef Y
#undef Z
#include "b_matrix_io.h"

#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_sort.h>
#include <gsl/gsl_sort_vector.h>

#define DEFAULT_BOOL -1

#define SQR2(x) ((x) * (x))

/* We output:                                                                 */
/*     chisq:                 Fit of tensor to data                           */
/*     lambda{1,2,3}:         Primary Axes (x,y,z) of tensor                  */
/*     theta, phi and gamma:  Euler angles of tensor (orientation)            */
/*     diffmatxx:             The complete Diffusion Matrix (possibly)        */
#define CHISQ_IDX 0
#define LAMBDA1_IDX 1
#define LAMBDA2_IDX 2
#define LAMBDA3_IDX 3
#define THETA_IDX 4
#define PHI_IDX 5
#define GAMMA_IDX 6

#define DIFFMAT00_IDX 7
#define DIFFMAT01_IDX 8
#define DIFFMAT02_IDX 9
#define DIFFMAT10_IDX 10
#define DIFFMAT11_IDX 11
#define DIFFMAT12_IDX 12
#define DIFFMAT20_IDX 13
#define DIFFMAT21_IDX 14
#define DIFFMAT22_IDX 15

char     outfile_names[][11] = {
   "chisq",
   "lambda1",
   "lambda2",
   "lambda3",
   "theta",
   "phi",
   "gamma",

   "diffmat00",
   "diffmat01",
   "diffmat02",
   "diffmat10",
   "diffmat11",
   "diffmat12",
   "diffmat20",
   "diffmat21",
   "diffmat22"
   };

/* Function prototypes */
void     do_math(void *caller_data, long num_voxels, int input_num_buffers,
                 int input_vector_length, double *input_data[], int output_num_buffers,
                 int output_vector_length, double *output_data[], Loop_Info * loop_info);
void     warp_B_mat(double **Def, double **B_I, B_Matrix * B);
void     print_version_info(void);

/* math data */
typedef struct {

   int      nbvals;
   int      ncoeff;

   /* SVD fitting stuff */
   double   svd_tol;
   gsl_matrix *svd_mat;
   gsl_matrix *svd_u;
   gsl_matrix *svd_v;
   gsl_vector *svd_s;
   gsl_vector *svd_x;
   gsl_vector *svd_b;
   double   chi_noise;

   /* diagonalisation workspace */
   gsl_matrix *diag_d;
   gsl_vector *diag_eig;
   gsl_matrix *diag_evor;
   gsl_eigen_symmv_workspace *diag_work;

   Volume   EPI_def_vol;
   Real     steps[MAX_VAR_DIMS];
   int      sizes[MAX_VAR_DIMS];

   int      output_diffmat;
   } Math_Data;

/* Argument variables */
int      verbose = FALSE;
int      quiet = FALSE;
int      clobber = FALSE;
int      max_buffer = 4 * 1024;
double   valid_range[2] = { 0.0, 0.0 };
int      copy_header = DEFAULT_BOOL;
int      is_signed = FALSE;
nc_type  dtype = NC_UNSPECIFIED;
char    *bmat_file = NULL;
int      output_diffmat = FALSE;
char    *EPI_file = NULL;
double   svd_tol = 1.0e-5;
double   chi_noise = 1.0;

/* Argument table */
ArgvInfo argTable[] = {
   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL, "General options:"},
   {"-version", ARGV_FUNC, (char *)print_version_info, (char *)NULL,
    "print version info and exit"},
   {"-verbose", ARGV_CONSTANT, (char *)TRUE, (char *)&verbose,
    "Print out extra information."},
   {"-quiet", ARGV_CONSTANT, (char *)TRUE, (char *)&quiet,
    "Be very quiet (I'm hunting wabbits)"},
   {"-clobber", ARGV_CONSTANT, (char *)TRUE, (char *)&clobber,
    "Clobber existing files."},
   {"-max_buffer", ARGV_INT, (char *)1, (char *)&max_buffer,
    "maximum size of buffers (in kbytes)"},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL, "\nOutfile Options:"},
   {"-copy_header", ARGV_CONSTANT, (char *)TRUE, (char *)&copy_header,
    "Copy header from 1st file (default)"},
   {"-nocopy_header", ARGV_CONSTANT, (char *)FALSE, (char *)&copy_header,
    "Don't copy header from 1st file."},
   {"-filetype", ARGV_CONSTANT, (char *)NC_UNSPECIFIED, (char *)&dtype,
    "Use data type of 1st file (default)"},
   {"-byte", ARGV_CONSTANT, (char *)NC_BYTE, (char *)&dtype,
    "Write out byte data."},
   {"-short", ARGV_CONSTANT, (char *)NC_SHORT, (char *)&dtype,
    "Write out short integer data."},
   {"-long", ARGV_CONSTANT, (char *)NC_LONG, (char *)&dtype,
    "Write out long integer data."},
   {"-float", ARGV_CONSTANT, (char *)NC_FLOAT, (char *)&dtype,
    "Write out single-precision data."},
   {"-double", ARGV_CONSTANT, (char *)NC_DOUBLE, (char *)&dtype,
    "Write out double-precision data."},
   {"-signed", ARGV_CONSTANT, (char *)TRUE, (char *)&is_signed,
    "Write signed integer data."},
   {"-unsigned", ARGV_CONSTANT, (char *)FALSE, (char *)&is_signed,
    "Write unsigned integer data."},
   {"-range", ARGV_FLOAT, (char *)2, (char *)valid_range,
    "Valid range for output data."},

   {NULL, ARGV_HELP, (char *)NULL, (char *)NULL, "\nDiffusion Options:"},
   {"-b_matrix", ARGV_STRING, (char *)1, (char *)&bmat_file,
    "File containing the B Matrix"},
   {"-output_diffmat", ARGV_CONSTANT, (char *)TRUE, (char *)&output_diffmat,
    "Output the complete Diffusion Matrix"},

   {"-EPI_def_grid", ARGV_STRING, (char *)1, (char *)&EPI_file,
    "Filename for the EPI distortions"},
   {"-svd_tolerance", ARGV_FLOAT, (char *)1, (char *)&svd_tol,
    "[0..1] Tolerance to use during SVD fitting"},
   {"-noise", ARGV_FLOAT, (char *)1, (char *)&chi_noise,
    "The variance of the signal in the background as used for chi-squared calculations"},

   {NULL, ARGV_HELP, NULL, NULL, ""},
   {NULL, ARGV_END, NULL, NULL, NULL}
   };

int main(int argc, char *argv[])
{
   char    *arg_string;
   int      i;
   double   fac;

   char   **infiles;
   char   **outfiles;
   int      n_infiles;
   int      n_outfiles;
   char    *outbase;
   B_Matrix *Bmat;
   Loop_Options *loop_opts;
   Math_Data math_data;

   /* Save time stamp and args */
   arg_string = time_stamp(argc, argv);

   /* Get arguments */
   if(ParseArgv(&argc, argv, argTable, 0) || (argc < 2)){
      print_error("\nUsage: %s [options] <infiles.mnc> ... <tensor_basename>\n", argv[0]);
      print_error("       %s -help\n\n", argv[0]);
      exit(EXIT_FAILURE);
      }

   /* set up in and outfile counts */
   n_infiles = argc - 2;
   infiles = &argv[1];
   n_outfiles = (output_diffmat) ? 16 : 7;
   outbase = argv[n_infiles + 1];

   /* check arguments */
   if(svd_tol > 1 || svd_tol < 0){
      fprintf(stderr, "%s: -svd_tol (%g) must be in the range 0..1\n\n", argv[0],
              svd_tol);
      exit(EXIT_FAILURE);
      }
   math_data.svd_tol = svd_tol;
   math_data.chi_noise = chi_noise;

   /* Initialise and read in the B_Matrix */
   if(access(bmat_file, F_OK) != 0){
      print_error("%s: Couldn't find B_matrix file: %s\n", argv[0], bmat_file);
      exit(EXIT_FAILURE);
      }

   Bmat = new_b_matrix(7, 0);
   if(input_b_matrix(bmat_file, Bmat) != OK){
      print_error("%s: Died whilst reading in B Matrix file: %s\n", argv[0], bmat_file);
      exit(EXIT_FAILURE);
      }

   /* check the infiles match and exist */
   if(n_infiles != Bmat->nbvals){
      print_error("%s: The number of infiles != Number of B Values\n", argv[0]);
      exit(EXIT_FAILURE);
      }

   if(verbose){
      fprintf(stderr, "\n==== Infiles ====\n");
      }
   for(i = 0; i < n_infiles; i++){
      if(verbose){
         fprintf(stderr, "bval_f[%02d]: %s\n", i, infiles[i]);
         }
      if(access(infiles[i], F_OK) != 0){
         print_error("%s: Couldn't find %s\n", argv[0], infiles[i]);
         exit(EXIT_FAILURE);
         }
      }

   /* Check for the EPI_distortions_file and open it (if doing this step) */
   if(EPI_file != NULL){
      print_error("%s: Distortion correction doesn't work yet!\n", argv[0]);
      exit(EXIT_FAILURE);
      }

   /* create and check for the outfiles */
   outfiles = (char **)malloc(sizeof(char *) * n_outfiles);
   if(verbose){
      fprintf(stderr, "\n==== Outfiles ====\n");
      }
   for(i = 0; i < n_outfiles; i++){
      outfiles[i] = (char *)malloc(sizeof(char) * (strlen(outbase) +
                                                   strlen(outfile_names[i]) + 5));
      sprintf(outfiles[i], "%s.%s.mnc", argv[n_infiles + 1], outfile_names[i]);
      if(verbose){
         fprintf(stderr, "tensor[%02d]: %s\n", i, outfiles[i]);
         }
      if(access(outfiles[i], F_OK) == 0 && !clobber){
         fprintf(stderr, "%s: %s exists, use -clobber to overwrite\n", argv[0],
                 outfiles[i]);
         exit(EXIT_FAILURE);
         }
      }

   /* alloc space for math_data structures */
   math_data.nbvals = Bmat->nbvals;
   math_data.ncoeff = Bmat->ncoeff;

   math_data.svd_mat = gsl_matrix_alloc(Bmat->nbvals, Bmat->ncoeff);
   math_data.svd_u = gsl_matrix_alloc(Bmat->nbvals, Bmat->ncoeff);
   math_data.svd_v = gsl_matrix_alloc(Bmat->ncoeff, Bmat->ncoeff);
   math_data.svd_s = gsl_vector_alloc(Bmat->ncoeff);
   math_data.svd_x = gsl_vector_alloc(Bmat->ncoeff);
   math_data.svd_b = gsl_vector_alloc(Bmat->nbvals);

   math_data.diag_d = gsl_matrix_alloc(3, 3);
   math_data.diag_eig = gsl_vector_alloc(3);
   math_data.diag_evor = gsl_matrix_alloc(3, 3);
   math_data.diag_work = gsl_eigen_symmv_alloc(3);

   math_data.output_diffmat = output_diffmat;

   /* form the B extended matrix (as a gsl matrix) */
   if(verbose){
      fprintf(stderr, "\n==== The b matrices (svd_u) ====\n");
      }
   for(i = 0; i < Bmat->nbvals; i++){
      gsl_matrix_set(math_data.svd_mat, i, 0, Bmat->B[i][0] * -1.0);
      gsl_matrix_set(math_data.svd_mat, i, 1, Bmat->B[i][1] * -1.0);
      gsl_matrix_set(math_data.svd_mat, i, 2, Bmat->B[i][2] * -1.0);
      gsl_matrix_set(math_data.svd_mat, i, 3, Bmat->B[i][3] * -2.0);
      gsl_matrix_set(math_data.svd_mat, i, 4, Bmat->B[i][4] * -2.0);
      gsl_matrix_set(math_data.svd_mat, i, 5, Bmat->B[i][5] * -2.0);
      gsl_matrix_set(math_data.svd_mat, i, 6, Bmat->B[i][6]);

      if(verbose){
         fprintf(stderr, "%02d  %12f %12f %12f %12f %12f %12f %2g\n",
                 i,
                 gsl_matrix_get(math_data.svd_mat, i, 0),
                 gsl_matrix_get(math_data.svd_mat, i, 1),
                 gsl_matrix_get(math_data.svd_mat, i, 2),
                 gsl_matrix_get(math_data.svd_mat, i, 3),
                 gsl_matrix_get(math_data.svd_mat, i, 4),
                 gsl_matrix_get(math_data.svd_mat, i, 5),
                 gsl_matrix_get(math_data.svd_mat, i, 6));
         }
      }

   /* copy svd_mat for chisq calculation */
   gsl_matrix_memcpy(math_data.svd_u, math_data.svd_mat);

   /* do svd decomposition and check tolerance */
   gsl_linalg_SV_decomp(math_data.svd_u, math_data.svd_v, math_data.svd_s,
                        math_data.svd_x);
   fac = math_data.svd_tol * gsl_vector_max(math_data.svd_s);
   for(i = 0; i < math_data.ncoeff; i++){
      if(gsl_vector_get(math_data.svd_s, i) < fac){
         gsl_vector_set(math_data.svd_s, i, 0.0);
         }
      }

   /* Do maths stuff */
   loop_opts = create_loop_options();
   set_loop_copy_all_header(loop_opts, copy_header);
   set_loop_verbose(loop_opts, (!quiet || verbose));
   set_loop_clobber(loop_opts, clobber);
   set_loop_datatype(loop_opts, dtype, is_signed, valid_range[0], valid_range[1]);
   set_loop_buffer_size(loop_opts, (long)1024 * max_buffer);

   voxel_loop(n_infiles, infiles, n_outfiles, outfiles, arg_string, loop_opts, do_math,
              (void *)&math_data);

   free_loop_options(loop_opts);

   /* be tidy */
   gsl_matrix_free(math_data.svd_mat);
   gsl_matrix_free(math_data.svd_u);
   gsl_matrix_free(math_data.svd_v);
   gsl_vector_free(math_data.svd_s);
   gsl_vector_free(math_data.svd_x);
   gsl_vector_free(math_data.svd_b);

   gsl_matrix_free(math_data.diag_d);
   gsl_vector_free(math_data.diag_eig);
   gsl_matrix_free(math_data.diag_evor);
   gsl_eigen_symmv_free(math_data.diag_work);

   return (EXIT_SUCCESS);
   }

void do_math(void *caller_data, long num_voxels, int input_num_buffers,
             int input_vector_length, double *input_data[], int output_num_buffers,
             int output_vector_length, double *output_data[], Loop_Info * loop_info)
{

   /* Get pointer to window info */
   Math_Data *md = (Math_Data *) caller_data;

   gsl_permutation *perm;
   long     ivox;
   int      i, j;

   /* for chisq fit */
   double   sum;

   /* alloc for permutation object */
   perm = gsl_permutation_alloc(3);

   for(ivox = 0; ivox < num_voxels * input_vector_length; ivox++){

      /* take log of input series of measurements */
      for(i = 0; i < md->nbvals; i++){
         if(input_data[i][ivox] < 1.0){
            gsl_vector_set(md->svd_b, i, 0.0);  /* natural log(1.0) = 0.0 */
            }
         else {
            gsl_vector_set(md->svd_b, i, log(input_data[i][ivox]));
            }
         }

      /* solve for this input series -- ax = b */
      gsl_linalg_SV_solve(md->svd_u, md->svd_v, md->svd_s, md->svd_b, md->svd_x);

      /* calculate chi-squared fit of this series to B-matrix (euclidian distance) */
      output_data[CHISQ_IDX][ivox] = 0.0;
      for(i = 0; i < md->nbvals; i++){
         sum = 0.0;
         for(j = 0; j < md->ncoeff; j++){
            sum += gsl_matrix_get(md->svd_mat, i, j) * gsl_vector_get(md->svd_x, j);
            }
         output_data[CHISQ_IDX][ivox] +=
            SQR2((gsl_vector_get(md->svd_b, i) - sum) / input_data[i][ivox])
            / md->chi_noise;
         }

      /* On exit from gsl_linalg_SV_solve svd_x contains, xx, yy, zz, xy, xz, yz, lnY(0) */
      gsl_matrix_set(md->diag_d, 0, 0, gsl_vector_get(md->svd_x, 0));
      gsl_matrix_set(md->diag_d, 1, 1, gsl_vector_get(md->svd_x, 1));
      gsl_matrix_set(md->diag_d, 2, 2, gsl_vector_get(md->svd_x, 2));

      /* off diagonal elements */
      gsl_matrix_set(md->diag_d, 0, 1, gsl_vector_get(md->svd_x, 3));
      gsl_matrix_set(md->diag_d, 1, 0, gsl_vector_get(md->svd_x, 3));
      gsl_matrix_set(md->diag_d, 0, 2, gsl_vector_get(md->svd_x, 4));
      gsl_matrix_set(md->diag_d, 2, 0, gsl_vector_get(md->svd_x, 4));
      gsl_matrix_set(md->diag_d, 1, 2, gsl_vector_get(md->svd_x, 5));
      gsl_matrix_set(md->diag_d, 2, 1, gsl_vector_get(md->svd_x, 5));

      /* output the diffusion matrix if required */
      if(md->output_diffmat){
         output_data[DIFFMAT00_IDX][ivox] = gsl_matrix_get(md->diag_d, 0, 0);
         output_data[DIFFMAT01_IDX][ivox] = gsl_matrix_get(md->diag_d, 0, 1);
         output_data[DIFFMAT02_IDX][ivox] = gsl_matrix_get(md->diag_d, 0, 2);

         output_data[DIFFMAT10_IDX][ivox] = gsl_matrix_get(md->diag_d, 1, 0);
         output_data[DIFFMAT11_IDX][ivox] = gsl_matrix_get(md->diag_d, 1, 1);
         output_data[DIFFMAT12_IDX][ivox] = gsl_matrix_get(md->diag_d, 1, 2);

         output_data[DIFFMAT20_IDX][ivox] = gsl_matrix_get(md->diag_d, 2, 0);
         output_data[DIFFMAT21_IDX][ivox] = gsl_matrix_get(md->diag_d, 2, 1);
         output_data[DIFFMAT22_IDX][ivox] = gsl_matrix_get(md->diag_d, 2, 2);
         }

      /* diagonalisation of the tensor */
      gsl_eigen_symmv(md->diag_d, md->diag_eig, md->diag_evor, md->diag_work);

      /* get permutation vector for eigen values and vectors */
      gsl_sort_vector_index(perm, md->diag_eig);

      output_data[LAMBDA1_IDX][ivox] =
         gsl_vector_get(md->diag_eig, gsl_permutation_get(perm, 2));
      output_data[LAMBDA2_IDX][ivox] =
         gsl_vector_get(md->diag_eig, gsl_permutation_get(perm, 1));
      output_data[LAMBDA3_IDX][ivox] =
         gsl_vector_get(md->diag_eig, gsl_permutation_get(perm, 0));

      /* convert the rotation matrix to euler angles */
      /* theta */
      output_data[THETA_IDX][ivox] =
         acos(gsl_matrix_get(md->diag_evor, gsl_permutation_get(perm, 2), 2));

      if(output_data[THETA_IDX][ivox] == 0.0){
         /* phi */
         output_data[PHI_IDX][ivox] = 0.0;

         /* gamma */
         output_data[GAMMA_IDX][ivox] =
            acos(gsl_matrix_get(md->diag_evor, gsl_permutation_get(perm, 0), 0));
         }
      else {
         /* phi */
         output_data[PHI_IDX][ivox] =
            asin(gsl_matrix_get(md->diag_evor, gsl_permutation_get(perm, 2), 0) /
                 sin(output_data[THETA_IDX][ivox]));

         /* gamma */
         output_data[GAMMA_IDX][ivox] =
            acos(gsl_matrix_get(md->diag_evor, gsl_permutation_get(perm, 1), 2) /
                 sin(output_data[THETA_IDX][ivox]));
         }
      }

   /* tidy up */
   gsl_permutation_free(perm);

   return;
   }

void print_version_info(void)
{
   fprintf(stdout, "\n");
   fprintf(stdout, "%s version %s\n", PACKAGE, VERSION);
   fprintf(stdout, "Comments to %s\n", PACKAGE_BUGREPORT);
   fprintf(stdout, "\n");
   exit(EXIT_SUCCESS);
   }
